//
//  GSFontMaster.h
//  Glyphs
//
//  Created by Georg Seifert on 6.9.07.
//  Copyright 2007 schriftgestaltung.de. All rights reserved.
//

#import <Cocoa/Cocoa.h>

@class GSFont;
@class GSGlyph;
@class GSLayer;
@class GSGuide;
@class GSCustomParameter;
@class GSAlignmentZone;
@class GSIndexedLookupTable;
@class GSFontInfoProperty;
@class GSInfoValue;
@class GSTTStem;
#import <GlyphsCore/GSAxis.h>
#import <GlyphsCore/GSContainerProtocol.h>
#import <GlyphsCore/GSMetric.h>
#import <GlyphsCore/GSMetricValue.h>
#import <GlyphsCore/GSUserDataProtocol.h>

/** The class defining the fontmaster object */
#ifndef GLYPHS_VIEWER
@interface GSFontMaster : NSObject <NSCoding, NSCopying, GSContainerProtocol, GSUserDataProtocol>
#else
@interface GSFontMaster : NSObject
#endif
{
	//NSInteger _index;
	GSFont *__unsafe_unretained _parent;
	NSMutableDictionary *_userData;
	NSString *_id;

	NSMutableDictionary *_changeCount;

	NSString *_name;
	NSString *_iconName;
	NSUInteger _gridMain;
	NSUInteger _gridSubDivision;

#pragma GSAutoCodeStart ivars
#ifndef GLYPHS_VIEWER
	NSMutableArray<GSGuide *> *_guides;
#endif
	NSMutableArray<GSCustomParameter *> *_customParameters;
	NSMutableArray<GSFontInfoProperty *> *_properties;

#pragma GSAutoCodeEnd ivars
}

@property (unsafe_unretained, nonatomic) GSFont *font;
/// @name Properties

/** the name of the FontMaster
e.g; "Regular", "Condensed", ...
*/

@property (nonatomic, strong) NSString *name;

@property (nonatomic, strong) NSString *iconName;

@property (nonatomic, readonly) NSImage *icon;

@property (nonatomic, readonly) NSString *nameUI;

@property (nonatomic, strong) NSDictionary *readBuffer;

@property (readonly) NSArray<GSMetricValue *> *axes;

@property (strong, nonatomic) GSAxisValues *axesValues;

@property (readonly, nonatomic) CGFloat *axesValuesList;

- (CGFloat)axisValueValueForId:(NSString *)aId;
- (void)setAxisValueValue:(CGFloat)value forId:(NSString *)aId;

- (void)setAxisValue:(GSMetricValue *)value forId:(NSString *)aId;

/// The id
@property (copy, nonatomic) NSString *id;

@property (nonatomic) BOOL visible;

/** Delayes redrawing and updates.
 
 Will record the changes and update once in `enableUpdateInterface`. If you don’t need updates, use `stopUpdateInterface`
 
 @see stopUpdateInterface
 @warning Always call enableUpdateInterface afterwards
 */
- (void)disableUpdateInterface;

/** Disable redrawing and updates completly.
 
 Changes will not trigger an update. Only use this calefully and call enableUpdateInterface if you need future updates.
 
 @see disableUpdateInterface
 @warning Always call enableUpdateInterface afterwards
 */
- (void)stopUpdateInterface;

/** Reanables redrawing and updates.
 
 This only has an effect if disableUpdateInterface was called before
 */
- (void)enableUpdateInterface;

- (void)enableFutureUpdates;
/** Check if the updating is enabled
 
 @return NO if disableUpdateInterface was called
 */
- (BOOL)isUpdateInterfaceEnabled;

#pragma mark - Metrics

- (void)setDefaultMetric:(CGFloat)value forType:(GSMetricsType)type;

- (GSMetricValue *)defaultMetricValueForType:(GSMetricsType)type;

- (CGFloat)defaultMetricForType:(GSMetricsType)type;

- (CGFloat)metricForType:(GSMetricsType)type layer:(GSLayer *)layer;

/** The default ascender.
 This is kept for compatibilit, for now.
 */
@property (nonatomic, readonly) CGFloat defaultAscender;

- (CGFloat)ascenderForLayer:(GSLayer *)layer;

@property (nonatomic, readonly) CGFloat defaultDescender;

- (CGFloat)descenderForLayer:(GSLayer *)layer;

/// The capHeight.
@property (nonatomic, readonly) CGFloat defaultCapHeight;

//- (CGFloat)capHeightForLayer:(GSLayer*)layer;
/// The xHeight.
@property (nonatomic, readonly) CGFloat defaultXHeight;

- (CGFloat)xHeightForLayer:(GSLayer *)layer;

- (CGFloat)slantHeightForLayer:(GSLayer *)layer;

- (CGFloat)topHeightForLayer:(GSLayer *)layer;

#pragma mark metricValues

@property (nonatomic, strong) NSDictionary<NSString *, GSMetricValue *> *metricValues;

- (GSMetricValue *)valueForMetric:(GSMetric *)metric;

- (void)setMetricValue:(GSMetricValue *)object forId:(NSString *)ID;

- (GSMetricValue *)setMetricPosition:(CGFloat)pos overshoot:(CGFloat)over type:(GSMetricsType)type name:(GSMetricsKey)name;

- (NSArray<GSMetricValue *> *)metrics;

#pragma mark StemValues

@property (nonatomic, strong) NSDictionary<NSString *, GSInfoValue *> *stemValues;

- (GSInfoValue *)valueForStem:(GSMetric *)stem;

- (CGFloat)valueValueForStemId:(NSString *)stemId;

- (void)setStemValue:(GSInfoValue *)object forId:(NSString *)ID;

- (void)setStemValueValue:(CGFloat)value forId:(NSString *)ID;

- (NSArray<GSInfoValue *> *)stems;

- (NSArray *)stemValuesArray;
/// convenient for TT export
- (NSArray<GSTTStem *> *)ttStems;

#pragma mark NumberValues

@property (nonatomic, strong) NSDictionary<NSString *, GSInfoValue *> *numberValues;

- (GSInfoValue *)numberValueForId:(NSString *)ID;

- (GSInfoValue *)numberValueForName:(NSString *)name;

- (CGFloat)numberValueValueForId:(NSString *)ID;

- (CGFloat)numberValueValueForName:(NSString *)name;

- (void)setNumberValue:(GSInfoValue *)object forId:(NSString *)ID;

- (void)setNumberValue:(GSInfoValue *)object forName:(NSString *)name;

- (void)setNumberValueValue:(CGFloat)floatValue forId:(NSString *)ID;

- (void)setNumberValueValue:(CGFloat)floatValue forName:(NSString *)name;

- (NSArray<GSInfoValue *> *)numbers;

#pragma mark -

/** The Italic angle
 */
@property (nonatomic) CGFloat defaultItalicAngle;

- (CGFloat)italicAngleForLayer:(GSLayer *)layer;

///// list of alignment zones
//@property (readonly, nonatomic) NSArray *alignmentZones;

- (instancetype)initFast;

//@property (strong, nonatomic) NSMutableArray *customParameters;

/// @name Initializing an Font Master

/**
 Inits with a dict.
 
 @param dict The property list representation
 @param formatVersion the version of the property list
 @return a fontmaster object
 */
- (instancetype)initWithDict:(NSDictionary *)dict format:(GSFormatVersion)formatVersion;

- (void)loadReadBuffer;

#ifndef GLYPHS_VIEWER

/** Returns the content of the object to store it in pList.
 
 This is used to store the data in the .glyphs file.
 @param format the version of the dict
 */
- (NSDictionary *)propertyListValueFormat:(GSFormatVersion)format;

- (BOOL)saveToFile:(FILE *)file format:(GSFormatVersion)formatVersion error:(NSError **)error;

- (NSArray *)blueValues;

- (NSArray *)otherBlues;

#endif

- (NSArray *)alignmentZonesForLayer:(GSLayer *)layer;

+ (int)isOnMetrics:(CGFloat)position metrics:(NSArray <GSMetricValue *>*)metrics;

- (BOOL)isOnMetricsLine:(NSPoint)position layer:(GSLayer *)layer;

@property (nonatomic, assign) NSUInteger gridMain;

@property (nonatomic, assign) NSUInteger gridSubDivision;

- (CGFloat)gridLength;

#ifndef GLYPHS_VIEWER

+ (CGFloat)localizedParameterValue:(NSString *)parameter glyph:(GSGlyph *)glyph;

+ (NSString *)localizedParameter:(NSString *)parameter glyph:(GSGlyph *)glyph;

- (CGFloat)localizedCustomValueForKey:(NSString *)key layer:(GSLayer *)layer;

- (CGFloat)localizedCustomValueForKey:(NSString *)key script:(NSString *)script;

/// @name Other Methods
/** Scales the Font Master
Scales the vertical metrics, alignment zones and guide lines
@param scale the scale factor (1 means no scaling.)
*/
- (void)scaleBy:(CGFloat)scale;

/**
 Adds a new font info value to the array specified by `key` or updates an existing one.
 
 @param name 		The key whose plural form specifies an array property of `GSFontInfoValue` objects on the receiver.
 @param value 		The actual value to be added.
 @param languageTag	The language under which the the value should be stored. Defaults to `English` if nil is passed.
 */
- (void)setProperty:(NSString *)name value:(NSString *)value languageTag:(nullable NSString *)languageTag;

- (GSFontInfoProperty *)propertyForName:(NSString *)name;

- (NSString *)defaultPropertyForName:(NSString *)name;

+ (NSArray *)iconNames;

- (void)invalidateCustomParameterCache;
#endif

- (void)setReadBuffer:(NSObject *)values forKey:(NSString *)key;

#pragma mark - start auto code

#pragma GSAutoCodeStart methods

#pragma mark Guides

/// The guides.
@property (nonatomic, strong) NSMutableArray<GSGuide *> *guides;

/** The number of guides */
- (NSUInteger)countOfGuides;

/** Returns object at idx in guides
 
 @param idx The index
 @return the object at index
 */
- (GSGuide *)objectInGuidesAtIndex:(NSUInteger)idx;

/** Returns the index of guide in guides
 
 @param guide the object
 @return the index of the object
 */
- (NSUInteger)indexOfObjectInGuides:(GSGuide *)guide;

/** Adds the guide
 
 @param guide the object to be added
 */
- (void)addGuide:(GSGuide *)guide;

/** Inserts the guide at index into guides
 
 @param guide The object to insert
 @param idx The index
 */
- (void)insertObject:(GSGuide *)guide inGuidesAtIndex:(NSUInteger)idx;

/** Removes the guide
 
 @param guide the object to be removed
 */
- (void)removeObjectFromGuides:(GSGuide *)guide;

/** Removes the guide at idx
 
 @param idx The index
 */
- (void)removeObjectFromGuidesAtIndex:(NSUInteger)idx;

#pragma mark CustomParameters

/// @name customParameters

/**
 The customParameters.
 A list of customParameters objects
 */
@property (nonatomic, strong) NSMutableArray<GSCustomParameter *> *customParameters;

/** The count of customParameters */
- (NSUInteger)countOfCustomParameters;

- (GSCustomParameter *)objectInCustomParametersAtIndex:(NSUInteger)idx;

- (BOOL)customBoolValueForKey:(NSString *)key;

- (id)customColorValueForKey:(NSString *)key;

/// The value of the customParameters where name == Key
- (id)customValueForKey:(NSString *)key;

/// The customParameters where name == Key
- (GSCustomParameter *)customParameterForKey:(NSString *)key;

/** Set the value for key
 
 It will look for an existing paramter with the name an overwrite its value, or adds a new parameter
 @param value The value to add
 @param key   The name of the paramter
 */
- (void)setCustomValue:(id)value forKey:(NSString *)key;

- (void)addCustomParameter:(GSCustomParameter *)property;

- (void)removeObjectFromCustomParameters:(GSCustomParameter *)customParameter;

/** Removes the first paramter with then Name
 
 @param key The name
 */
- (void)removeObjectFromCustomParametersForKey:(NSString *)key;

- (void)insertObject:(GSCustomParameter *)customParameter inCustomParametersAtIndex:(NSUInteger)idx;

- (void)removeObjectFromCustomParametersAtIndex:(NSUInteger)idx;

- (void)replaceObjectInCustomParametersAtIndex:(NSUInteger)idx withObject:(GSCustomParameter *)customParameter;

#pragma mark Properties
/// @name Properties

/** The properties.
 this contains (localisable) names and other font info settings
 */
@property (nonatomic, strong) NSMutableArray<GSFontInfoProperty *> *properties;

/** The number of properties */
- (NSUInteger)countOfProperties;

/** Returns object at idx in properties
 
 @param idx The index
 @return the object at index
 */
- (GSFontInfoProperty *)objectInPropertiesAtIndex:(NSUInteger)idx;

/** Returns the index of property in properties
 
 @param property the object
 @return the index of the object
 */
- (NSUInteger)indexOfObjectInProperties:(GSFontInfoProperty *)property;

/** Adds the property
 
 @param property the object to be added
 */
- (void)addProperty:(GSFontInfoProperty *)property;

/** Inserts the property at index into properties
 
 @param property The object to insert
 @param idx The index
 */
- (void)insertObject:(GSFontInfoProperty *)property inPropertiesAtIndex:(NSUInteger)idx;

/** Removes the property
 
 @param property the object to be removed
 */
- (void)removeObjectFromProperties:(GSFontInfoProperty *)property;

/** Removes the property at idx
 
 @param idx The index
 */
- (void)removeObjectFromPropertiesAtIndex:(NSUInteger)idx;

#pragma mark UserData

/// @name UserData
/** Place to store data.
 
 Here it is possible to store something. Please use a unique key.
 The objects should be able to serialize to a plist. If they are not, they should return all infos to store in description (as String).
 */
@property (strong, nonatomic, nullable) NSDictionary *userData;


/**
 convenience accessor to get to the content of the userData dict
 
 @param key the key
 @return the data stored with key
 */
- (NSObject * _Nullable)userDataForKey:(NSString *)key;


/** Adds something to the fonts userData.
 
 This also triggers undo/document dirty state.
 
 @param value The object should be able to serialize to a plist. If they are not, they should return all infos to store in description (as String).
 @param key   Please use an uniqe key that gives some indecation who put it there (e.g. prefix it with your name or project).
 */
- (void)setUserData:(nullable id)value forKey:(nonnull NSString *)key;


/** Removed the object with this key.
 
 @param key The key to remove
 */
- (void)removeUserDataForKey:(NSString *)key;
#pragma GSAutoCodeEnd methods

@end
